/**
  * File: DomainsFactory.java
  * Content: Class to send DoaminGroupHandler messages.
  * Author: LuisM Pena
  * Date: 03rd December 2000
  * Version: 0.51.00
  * Last change:
  *
  **/

package sensei.domains;

import java.io.IOException;

import sensei.middleware.domains.BehaviourOnViewChanges;
import sensei.middleware.domains.DynamicSubgroupInfo;
import sensei.middleware.domains.EVERY_SUBGROUP;
import sensei.middleware.domains.EndOfTransactionMessage;
import sensei.middleware.domains.MemberProperties;
import sensei.middleware.domains.MessageC;
import sensei.middleware.domains.MessageCreateSubgroup;
import sensei.middleware.domains.MessageE;
import sensei.middleware.domains.MessageMemberProperties;
import sensei.middleware.domains.MessageP;
import sensei.middleware.domains.MessageRemoveSubgroup;
import sensei.middleware.domains.MessageS;
import sensei.middleware.domains.Property;
import sensei.middleware.domains.ReplicaInformation;
import sensei.middleware.domains.SubgroupCoordination;
import sensei.middleware.domains.SubgroupProperties;
import sensei.middleware.domains.SubgroupState;
import sensei.middleware.gms.GroupHandler;
import sensei.middleware.gms.Message;

/**
  * Class to send DoaminGroupHandler messages.
  **/
class DomainsFactory {

	//*************************************************************************************//
	//**************************** CONSTRUCTOR ********************************************//
	//*************************************************************************************//

	/**
	 * The constructor creates one instance for each type of message, as only one message is indeed sent each
	 * time (a messageP cannot be sent on the same round to two different members, for example)
	 */
	public DomainsFactory(GroupHandler group, BehaviourOnViewChanges behaviour, boolean coordinatorElectorRegistered, boolean propertiesAllowed, boolean dynamicSubgroupsAllowed) {
		Trace.code("DomainsFactory.java -> DomainsFactory ( GroupHandler group , BehaviourOnViewChanges behaviour , boolean coordinatorElectorRegistered , boolean propertiesAllowed , boolean dynamicSubgroupsAllowed )");
		this.group = group;
		messageP = new MessagePImpl(behaviour, coordinatorElectorRegistered, propertiesAllowed, dynamicSubgroupsAllowed);
	}

	/**
	 * Releases resources
	 */
	public void freeResources() {
		Trace.code("DomainsFactory.java -> void freeResources ( )");
		group = null;
		messageP = null;
	}

	/**
	 * Returns true if the message is sent on this view
	 */
	public boolean sendMessageP(int groupMember, int viewId, int nextDynamicSubgroupId, ReplicaInformation[] replicasInfo, MemberProperties[] groupProperties, SubgroupProperties[] subgroupsInformation) throws IOException {
		Trace.code("DomainsFactory.java -> boolean sendMessageP ( int groupMember , int viewId , int nextDynamicSubgroupId , ReplicaInformation [ ] replicasInfo , MemberProperties [ ] groupProperties , SubgroupProperties [ ] subgroupsInformation )");
		assert group != null;
		messageP.setInfo(viewId, nextDynamicSubgroupId, replicasInfo, groupProperties, subgroupsInformation);
		return group.sendMessage(groupMember, messageP);
	}

	public boolean sendMessageC(int groupMember, int viewId, SubgroupCoordination coordinations[], boolean containOwnProperties, Property properties[]) throws IOException {
		Trace.code("DomainsFactory.java -> boolean sendMessageC ( int groupMember , int viewId , SubgroupCoordination coordinations [ ] , boolean containOwnProperties , Property properties [ ] )");
		assert group != null;
		return group.sendMessage(groupMember, new MessageCImpl(viewId, coordinations, containOwnProperties, properties));
	}

	public boolean castMessageE(int viewId, int member, Property properties[]) throws IOException {
		Trace.code("DomainsFactory.java -> boolean castMessageE ( int viewId , int member , Property properties [ ] )");
		assert group != null;
		return group.castMessage(new MessageEImpl(viewId, member, properties));
	}

	public boolean sendMessageS(int groupMember, int viewId, SubgroupState subgoupStates[]) throws IOException {
		Trace.code("DomainsFactory.java -> boolean sendMessageS ( int groupMember , int viewId , SubgroupState subgoupStates [ ] )");
		assert group != null;
		return group.sendMessage(groupMember, new MessageSImpl(viewId, subgoupStates));
	}

	public boolean castMessageMemberProperties(Property[] properties) throws IOException {
		Trace.code("DomainsFactory.java -> boolean castMessageMemberProperties ( Property [ ] properties )");
		assert group != null;
		return group.castMessage(new MessageMemberPropertiesImpl(properties));
	}

	public boolean castMessageCreateSubgroup(DynamicSubgroupInfo info, int creatorId, boolean waitReception) throws IOException {
		Trace.code("DomainsFactory.java -> boolean castMessageCreateSubgroup ( DynamicSubgroupInfo info , int creatorId , boolean waitReception )");
		assert group != null;
		return group.castMessage(new MessageCreateSubgroupImpl(info, creatorId, waitReception));
	}

	public boolean castMessageRemoveSubgroup(int subgroup, DynamicSubgroupInfo info) throws IOException {
		Trace.code("DomainsFactory.java -> boolean castMessageRemoveSubgroup ( int subgroup , DynamicSubgroupInfo info )");
		assert group != null;
		return group.castMessage(new MessageRemoveSubgroupImpl(subgroup, info));
	}

	public boolean castEndOfTransaction(Message messages[]) throws IOException {
		Trace.code("DomainsFactory.java -> boolean castEndOfTransaction ( Message messages [ ] )");
		assert group != null;
		return group.castMessage(new EndOfTransactionMessageImpl(messages));
	}

	//*************************************************************************************//
	//**************************** INNER CLASSES *******************************************//
	//*************************************************************************************//

	static class MessagePImpl extends MessageP {
		public MessagePImpl(BehaviourOnViewChanges behaviour, boolean coordinatorElectorRegistered, boolean propertiesAllowed, boolean dynamicSubgroupsAllowed) {
			this.behaviour = behaviour;
			this.coordinatorElectorRegistered = coordinatorElectorRegistered;
			this.dynamicSubgroupsAllowed = dynamicSubgroupsAllowed;
			this.props = null;
			this.subgroupsInformation = null;
			this.replicasInfo = null;
			this.viewId = 0;
			this.unqueuedOnST = true;
			this.waitReception = false;
			this.subgroup = EVERY_SUBGROUP.value;
		}

		public void setInfo(int viewId, int nextDynamicSubgroupId, ReplicaInformation[] replicasInfo, MemberProperties[] properties, SubgroupProperties[] subgroupsInformation) {
			this.viewId = viewId;
			this.nextDynamicSubgroupId = nextDynamicSubgroupId;
			this.replicasInfo = replicasInfo;
			this.props = properties;
			this.subgroupsInformation = subgroupsInformation;
		}
	}

	static class MessageCImpl extends MessageC {
		public MessageCImpl(int viewId, SubgroupCoordination coordinations[], boolean containOwnProperties, Property properties[]) {
			this.viewId = viewId;
			this.coordinations = coordinations;
			this.containOwnProperties = containOwnProperties;
			this.props = properties;
			this.unqueuedOnST = true;
			this.waitReception = false;
			this.subgroup = EVERY_SUBGROUP.value;
		}
	}

	static class MessageEImpl extends MessageE {
		public MessageEImpl(int viewId, int member, Property properties[]) {
			this.viewId = viewId;
			this.member = member;
			this.props = properties;
			this.unqueuedOnST = true;
			this.waitReception = false;
			this.subgroup = EVERY_SUBGROUP.value;
		}
	}

	static class MessageSImpl extends MessageS {
		public MessageSImpl(int viewId, SubgroupState states[]) {
			this.viewId = viewId;
			this.states = states;
			this.unqueuedOnST = true;
			this.waitReception = false;
			this.subgroup = EVERY_SUBGROUP.value;
		}
	}

	static class MessageMemberPropertiesImpl extends MessageMemberProperties {
		public MessageMemberPropertiesImpl(Property[] properties) {
			this.props = properties;
			this.unqueuedOnST = true;
			this.waitReception = false;
			this.subgroup = EVERY_SUBGROUP.value;
		}
	}

	static class MessageCreateSubgroupImpl extends MessageCreateSubgroup {
		public MessageCreateSubgroupImpl(DynamicSubgroupInfo info, int creatorId, boolean waitReception) {
			this.info = info;
			this.creatorId = creatorId;
			this.unqueuedOnST = true;
			this.waitReception = waitReception;
			this.subgroup = EVERY_SUBGROUP.value;
		}
	}

	static class MessageRemoveSubgroupImpl extends MessageRemoveSubgroup {
		public MessageRemoveSubgroupImpl(int subgroup, DynamicSubgroupInfo info) {
			this.info = info;
			this.subgroupToRemove = subgroup;
			this.unqueuedOnST = true;
			this.subgroup = EVERY_SUBGROUP.value;
		}
	}

	static class EndOfTransactionMessageImpl extends EndOfTransactionMessage {
		public EndOfTransactionMessageImpl(Message messages[]) {
			this.actions = messages;
			this.unqueuedOnST = true;
			this.waitReception = false;
			this.unTransactionable = false; //just highlight it!
			this.subgroup = EVERY_SUBGROUP.value;
		}
	}

	//*************************************************************************************//
	//**************************** DATA MEMBERS *******************************************//
	//*************************************************************************************//

	GroupHandler group;
	MessagePImpl messageP;
}
