/**
  * File: MessageSempahore.java
  * Content: Implementation of a Semaphore
  * Author: LuisM Pena
  * Date: 30th September 2000
  * Last change:
  * Version: 0.51.00
  *
  **/

package sensei.domains;

import java.util.HashMap;
import java.util.Map;

/**
  * Class implementing a sempahore to use on the messages reception.
  * It defines the following protocol:
  *   A thread creates a semaphore entry, and another one releases it.
  *   The thread that creates the entry can perform 3 operations:
  *   -Create the entry, that can be joinable or not.
  *   -Join the entry, only if it's joinable
  *   -Remove the entry.
  *   The second thread can only perform the operation Release.
  *   two scenarios are possible:
  *   1- A- Entry is created as owned
  *      B- The same thread performs a join on the entry, which blocks.
  *      C- The second thread performs a release on the entry, that also blocks, but wakes up the first thread
  *      D- The creator thread removes the entry, waking up to the second thread, which receives a null entry,
  *         as it has been removed
  *   2- A- Entry is created as not owned
  *      B- The same thread joins the entry, blocking
  *      C- The second thread releases the entry, receiving it.
  *      D- The second thread removes the entry, waking up the first thread.
  *
  *   A join on the entry returns always true, except when the MessageSemaphore is destroyed, when false is returned
  *
  **/
public class MessageSemaphore {

	public class Entry {
		int id;
		boolean owned, joined, released, removed;
		Entry(int id, boolean owned) {
			this.id = id;
			this.owned = owned;
		}
		public int hashCode() {
			return id;
		}
		public boolean equals(Object other) {
			return (other instanceof Entry) && (((Entry) other).id == id);
		}
	}

	public MessageSemaphore() {
		Trace.code("MessageSemaphore.java -> MessageSemaphore ( )");
		entries = new HashMap();
		opt = new Entry(0, false);
		notReleased = true;
	}

	/**
	 * Frees any resources, notifying any pending 'join' calls
	 */
	public synchronized void freeResources() {
		Trace.code("MessageSemaphore.java -> synchronized void freeResources ( )");
		Trace.semaphoresReleased();
		entries.clear();
		notReleased = false;
		notifyAll();
	}

	public synchronized Entry create(int id, boolean owned) {
		Trace.code("MessageSemaphore.java -> synchronized Entry create ( int id , boolean owned )");
		Entry ret = null;
		if (notReleased) {
			ret = new Entry(id, owned);
			Object check = entries.put(ret, ret);
			assert check == null;
		}
		Trace.semaphoreCreated(id, owned);
		return ret;
	}

	public synchronized boolean join(Entry entry) {
		Trace.code("MessageSemaphore.java -> synchronized boolean join ( Entry entry )");
		Trace.semaphoreJoining(entry.id);
		if (notReleased) {
			assert !entry.joined;
			entry.joined = true;
			notifyAll();
			//if owned, wait until release is true
			//if not owned, wait until release and remove are true
			while (notReleased && (!entry.released || (!entry.owned && !entry.removed)))
				try {
					wait();
				}
				catch (InterruptedException ex) {
				}
		}
		Trace.semaphoreJoined(entry.id);
		return notReleased;
	}

	public synchronized Entry release(int id) {
		Trace.code("MessageSemaphore.java -> synchronized Entry release ( int id )");
		Entry ret = null;
		Trace.semaphoreReleasing(id);
		if (notReleased) {
			opt.id = id;
			ret = (Entry) entries.get(opt);
			assert ret != null;
			ret.released = true;
			notifyAll();
			//if owned, wait until remove is true
			//if not owned, wait until joined is true
			while (notReleased && ((ret.owned && !ret.removed) || (!ret.owned && !ret.joined)))
				try {
					wait();
				}
				catch (InterruptedException ex) {
				}

			if (ret.removed)
				ret = null;
		}
		Trace.semaphoreReleased(id);
		return ret;
	}

	public synchronized boolean remove(Entry entry) {
		Trace.code("MessageSemaphore.java -> synchronized boolean remove ( Entry entry )");
		if (notReleased) {
			assert entry.joined && entry.released;
			entry.removed = true;
			notifyAll();
			Object checker = entries.remove(entry);
			assert checker == entry;
		}
		Trace.semaphoreRemoved(entry.id);
		return notReleased;
	}

	boolean notReleased;
	Entry opt;
	Map entries; //key=value!
}
