/**
  * File: PropertyManager.java
  * Content: Helper class supporting the PropertyHandler interface to support the DomainGroupHandlerImpl
  * Author: LuisM Pena
  * Date: 03rd October 2000
  * Version: 0.51.00
  * Last change:
  *
  *
  **/

package sensei.domains;

import sensei.middleware.domains.MemberStateException;
import sensei.middleware.domains.MemberStateExceptionReason;
import sensei.middleware.domains.MemberProperties;
import sensei.middleware.domains.Property;

import sensei.util.SenseiError;
import sensei.util.IntWrapper;

import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
  * Helper class supporting the PropertyHandler interface to support the DomainGroupHandlerImpl
  **/
public class PropertyManager {

	//*************************************************************************************
	//**************************** CONSTRUCTOR ********************************************
	//*************************************************************************************

	public PropertyManager() {
		Trace.code("PropertyManager.java -> PropertyManager ( )");
		properties = new HashMap();
		wrapper = new IntWrapper();
		thisMember = new IntWrapper(Consts.INVALID_MEMBER_ID);
		setProperties(thisMember, new Property[0], false);
		messenger = null;
		ownPropertiesChanged = false;
		groupPropertiesKnown = false;
	}

	//*************************************************************************************
	//**************************** FREE RESOURCES *****************************************
	//*************************************************************************************
	public synchronized void freeResources() {
		Trace.code("PropertyManager.java -> synchronized void freeResources ( )");
		if (properties != null) {
			properties.clear();
			properties = null;
			wrapper = null;
			messenger = null;
		}
	}

	//*************************************************************************************
	//**************************** SET MEMBER ID ******************************************
	//*************************************************************************************

	public void setMemberId(int memberId) {
		Trace.code("PropertyManager.java -> void setMemberId ( int memberId )");
		assert properties != null;
		assert thisMember.i == Consts.INVALID_MEMBER_ID;
		//changing the key
		IntWrapper old = thisMember;
		thisMember = new IntWrapper(memberId);
		Object check = properties.put(thisMember, properties.remove(old));

		assert check == null;
	}

	//*************************************************************************************
	//**************************** SET MESSENGER ******************************************
	//*************************************************************************************

	public void setMessenger(DomainsFactory messenger) {
		Trace.code("PropertyManager.java -> void setMessenger ( DomainsFactory messenger )");
		assert properties != null;
		assert this.messenger == null;
		this.messenger = messenger;
		if (ownPropertiesChanged) {
			Trace.castMessageMemberProperties();
			castProperties(getProperties());
		}
	}

	void castProperties(Property[] properties) {
		Trace.code("PropertyManager.java -> void castProperties ( Property [ ] properties )");
		try {
			messenger.castMessageMemberProperties(properties);
		}
		catch (IOException ex) {
			SenseiError.unhandledException(Consts.AREA, ex); //Remote/CORBA exception, menaingless as the access is local!!!
		}
		catch (RuntimeException ex) {
			SenseiError.unhandledException(Consts.AREA, ex); //Remote/CORBA exception, menaingless as the access is local!!!
		}
	}

	//*************************************************************************************
	//**************************** CHECK PROPERTIES CHANGES *******************************
	//*************************************************************************************

	/**
	 * This method is provided for atomic operations. It resets the 'changed' field,
	 * and returns the properties only if they have changed. If the parameter getAnyway is true,
	 * the 'changed' field is still changed and the properties are returned anyway
	 */
	public synchronized PropertiesAndChange getOwnProperties(boolean getAnyway) {
		Trace.code("PropertyManager.java -> synchronized PropertiesAndChange getOwnProperties ( boolean getAnyway )");
		assert properties != null;
		boolean changed = ownPropertiesChanged;
		ownPropertiesChanged = false;
		PropertiesAndChange ret;
		if (getAnyway || changed)
			ret = new PropertiesAndChange(getProperties(), true);
		else
			ret = new PropertiesAndChange(new Property[0], false);
		return ret;
	}

	//*************************************************************************************
	//**************************** GROUP PROPERTIES OPERATIONS ****************************
	//*************************************************************************************

	/**
	 * Returns true if the properties from other group members have been received
	 */
	public synchronized boolean hasGroupProperties() {
		Trace.code("PropertyManager.java -> synchronized boolean hasGroupProperties ( )");
		return groupPropertiesKnown;
	}

	/**
	 * Checks that the member knows the group properties, otherwise it raises a MemberStateException
	 */
	void checkGroupPropertiesAreKnown() throws MemberStateException {
		Trace.code("PropertyManager.java -> void checkGroupPropertiesAreKnown ( )");
		if (!groupPropertiesKnown)
			throw new MemberStateException(MemberStateExceptionReason.MemberWithoutState);
	}

	public synchronized MemberProperties[] getGroupProperties() {
		Trace.code("PropertyManager.java -> synchronized MemberProperties [ ] getGroupProperties ( )");
		assert properties != null;
		assert hasGroupProperties();
		return getAllTheProperties();
	}

	/**
	 * Sets the group properties. Only those members that are not yet included are updated
	 */
	public synchronized void setGroupProperties(MemberProperties[] groupProperties) {
		Trace.code("PropertyManager.java -> synchronized void setGroupProperties ( MemberProperties [ ] groupProperties )");
		assert properties != null;
		assert !hasGroupProperties();
		groupPropertiesKnown = true;
		int size = groupProperties.length;
		while (size-- > 0) {
			MemberProperties memberProperties = groupProperties[size];
			wrapper.i = memberProperties.member;
			setProperties(wrapper, memberProperties.props, false);
		}
	}

	/**
	 * Sets the group properties without adding any. This is to be used when a member assumes state
	 * Returns false if it is not possible to receive the group properties, because they have
	 * been already received
	 */
	public synchronized boolean setGroupProperties() {
		Trace.code("PropertyManager.java -> synchronized boolean setGroupProperties ( )");
		assert properties != null;
		boolean ret = !hasGroupProperties();
		if (ret) {
			groupPropertiesKnown = true;
		}
		return ret;
	}

	/**
	 * Resets the group properties
	 */
	public synchronized void resetGroupProperties() {
		Trace.code("PropertyManager.java -> synchronized void resetGroupProperties ( )");
		assert properties != null;
		groupPropertiesKnown = false;
		Object thisProperties = properties.get(thisMember);
		properties.clear();
		properties.put(thisMember, thisProperties);
	}

	//*************************************************************************************
	//**************************** PROPERTY HANDLER INTERFACE *****************************
	//*************************************************************************************

	public synchronized MemberProperties[] getAllProperties() throws MemberStateException {
		Trace.code("PropertyManager.java -> synchronized MemberProperties [ ] getAllProperties ( )");
		assert properties != null;
		checkGroupPropertiesAreKnown();
		return getAllTheProperties();
	}

	public synchronized MemberProperties[] getPropertyForAllMembers(String property) throws MemberStateException {
		Trace.code("PropertyManager.java -> synchronized MemberProperties [ ] getPropertyForAllMembers ( String property )");
		assert properties != null;
		checkGroupPropertiesAreKnown();
		int size = properties.size();
		MemberProperties[] ret = new MemberProperties[size];
		Iterator it = properties.entrySet().iterator();
		while (it.hasNext()) {
			assert size > 0;
			Map.Entry entry = (Map.Entry) (it.next());
			IntWrapper key = (IntWrapper) (entry.getKey());
			//Map value = (Map) (entry.getValue());
			Property[] one = new Property[1];
			one[0] = new Property(property, getProperty(key, property));
			ret[--size] = new MemberProperties(key.i, one);
		}
		assert size == 0;
		return ret;
	}

	public synchronized Property[] getMemberProperties(int loc) throws MemberStateException {
		Trace.code("PropertyManager.java -> synchronized Property [ ] getMemberProperties ( int loc )");
		assert properties != null;
		checkGroupPropertiesAreKnown();
		wrapper.i = loc;
		return getProperties(wrapper);
	}

	public synchronized String getMemberProperty(int loc, String n) throws MemberStateException {
		Trace.code("PropertyManager.java -> synchronized String getMemberProperty ( int loc , String n )");
		assert properties != null;
		checkGroupPropertiesAreKnown();
		wrapper.i = loc;
		return getProperty(wrapper, n);
	}

	public synchronized Property[] getProperties() {
		Trace.code("PropertyManager.java -> synchronized Property [ ] getProperties ( )");
		assert properties != null;
		return getProperties(thisMember);
	}

	public synchronized String getProperty(String n) {
		Trace.code("PropertyManager.java -> synchronized String getProperty ( String n )");
		assert properties != null;
		return getProperty(thisMember, n);
	}

	public synchronized void setProperties(Property[] setProps) {
		Trace.code("PropertyManager.java -> synchronized void setProperties ( Property [ ] setProps )");
		assert properties != null;
		if (messenger == null) {
			//if there is no messenger, properties are directly updated
			ownPropertiesChanged = true;
			boolean ok = setProperties(thisMember, setProps, true);
			assert ok;
		}
		else {
			//With a messenger, properties must be updated when they're received, with the other members.
			//They are updated in parallel on a private Map, and then sent
			Trace.castMessageMemberProperties();
			castProperties(setProps);
		}
	}

	public synchronized void addProperties(Property[] addProps) {
		Trace.code("PropertyManager.java -> synchronized void addProperties ( Property [ ] addProps )");
		assert properties != null;
		if (messenger == null) {
			//if there is no messenger, properties are directly updated
			ownPropertiesChanged = true;
			addProperties((Map) (properties.get(thisMember)), addProps);
		}
		else {
			//With a messenger, properties must be updated when they're received, with the other members.
			//They are updated in parallel on a private Map, and then sent
			HashMap copied = new HashMap((Map) (properties.get(thisMember)));
			addProperties(copied, addProps);
			Property[] finalProps = getProperties(copied);
			castProperties(finalProps);
			Trace.castMessageMemberProperties();
		}
	}

	public synchronized void removeProperties(Property[] remProps) {
		Trace.code("PropertyManager.java -> synchronized void removeProperties ( Property [ ] remProps )");
		assert properties != null;
		if (messenger == null) {
			//if there is no messenger, properties are directly updated
			ownPropertiesChanged = true;
			removeProperties((Map) (properties.get(thisMember)), remProps);
		}
		else {
			//With a messenger, properties must be updated when they're received, with the other members.
			//They are updated in parallel on a private Map, and then sent
			HashMap copied = new HashMap((Map) (properties.get(thisMember)));
			removeProperties(copied, remProps);
			Property[] finalProps = getProperties(copied);
			castProperties(finalProps);
			Trace.castMessageMemberProperties();
		}
	}

	//*************************************************************************************
	//**************************** OPERATIONS TO UPDATE OTHER's PROPERTIES ****************
	//*************************************************************************************

	/**
	 * Sets a member's properties. The member and must be already existing
	 * when newMember is false (otherwisse, it must no exist).
	 * If the member set is this member, some special behaviour is required:
	 * 1- If the messenger is already set, it updates the properties.
	 * 2- If it is not set, it does nothing
	 */
	public synchronized boolean setProperties(int member, Property[] setProps, boolean newMember) {
		Trace.code("PropertyManager.java -> synchronized boolean setProperties ( int member , Property [ ] setProps , boolean newMember )");
		assert properties != null;
		boolean ret;
		if (member == thisMember.i)
			if (messenger == null)
				ret = true;
			else
				ret = setProperties(thisMember, setProps, true);
		else {
			wrapper.i = member;
			ret = setProperties(wrapper, setProps, !newMember);
		}
		return ret;
	}

	//*************************************************************************************
	//**************************** HELPER METHODS *****************************************
	//*************************************************************************************

	private MemberProperties[] getAllTheProperties() {
		Trace.code("PropertyManager.java -> MemberProperties [ ] getAllTheProperties ( )");
		int size = properties.size();
		MemberProperties[] ret = new MemberProperties[size];
		Iterator it = properties.entrySet().iterator();
		while (it.hasNext()) {
			assert size > 0;
			Map.Entry entry = (Map.Entry) (it.next());
			IntWrapper key = (IntWrapper) (entry.getKey());
			Map value = (Map) (entry.getValue());
			ret[--size] = new MemberProperties(key.i, getProperties(value));
		}
		assert size == 0;
		return ret;
	}

	private Property[] getProperties(IntWrapper wrapper) {
		Trace.code("PropertyManager.java -> Property [ ] getProperties ( IntWrapper wrapper )");
		return getProperties((Map) (properties.get(wrapper)));
	}

	/**
	 * convert the map (string, value) into Property[]
	 */
	private Property[] getProperties(Map props) {
		Trace.code("PropertyManager.java -> Property [ ] getProperties ( Map props )");
		Property[] ret;
		if (props == null)
			ret = new Property[0];
		else {
			int size = props.size();
			ret = new Property[size];
			Iterator it = props.entrySet().iterator();
			while (it.hasNext()) {
				assert size > 0;
				Map.Entry entry = (Map.Entry) (it.next());
				ret[--size] = new Property((String) (entry.getKey()), (String) (entry.getValue()));
			}
			assert size == 0;
		}
		return ret;
	}

	private String getProperty(IntWrapper wrapper, String n) {
		Trace.code("PropertyManager.java -> String getProperty ( IntWrapper wrapper , String n )");
		String ret = null;
		HashMap props = (HashMap) (properties.get(wrapper));
		if (props != null)
			ret = (String) (props.get(n));
		return ret;
	}

	/**
	 * Returns true if the set has been all right: false means that the member had no properties and
	 * existing was true, or the member had properties and existing was false
	 */
	private boolean setProperties(IntWrapper member, Property[] setProps, boolean existing) {
		Trace.code("PropertyManager.java -> boolean setProperties ( IntWrapper member , Property [ ] setProps , boolean existing )");
		boolean ret;
		Map props = (Map) (properties.get(member));
		if (props == null) {
			ret = !existing;
			if (ret) {
				props = new HashMap();
				properties.put(new IntWrapper(member.i), props);
			}
		}
		else {
			ret = existing;
			if (ret)
				props.clear();
		}
		if (ret)
			addProperties(props, setProps);
		return ret;
	}

	private void addProperties(Map props, Property[] addProps) {
		Trace.code("PropertyManager.java -> void addProperties ( Map props , Property [ ] addProps )");
		int size = addProps.length;
		while (size-- > 0) {
			Property prop = addProps[size];
			props.put(prop.nam, prop.val);
		}
	}

	private void removeProperties(Map props, Property[] remProps) {
		Trace.code("PropertyManager.java -> void removeProperties ( Map props , Property [ ] remProps )");
		int size = remProps.length;
		while (size-- > 0)
			props.remove(remProps[size].nam);
	}

	//*************************************************************************************
	//**************************** DATA MEMBERS *******************************************
	//*************************************************************************************

	//data is kept in a HashMap, a IntWrapper identifies the members, and the value is
	//another HashMap with the properties for that member (String/String)
	Map properties;

	//The PropertyManager automatically sends the properties when it has a messenger. If it has no
	//messenger, it keeps track of changes on the own properties.
	DomainsFactory messenger;

	//IntWrappers to access quicker the data
	IntWrapper wrapper;
	IntWrapper thisMember; //this member

	//ownPropertiesChanged is true if the properties for this member have changed. Has no sense when the
	//messenger is set, as any changes are directly propagated
	boolean ownPropertiesChanged;

	//groupPropertiesKnown is true when the PropertyManager has received the properties of other group's members.
	boolean groupPropertiesKnown;
}
