/**
  * File: TransactionsManager.java
  * Content: Class supporting the use of transactions
  * Author: LuisM Pena
  * Date: 09th September 2001
  * Version: 0.51.00
  * Last change:
  *
  *
  **/

package sensei.domains;

import sensei.middleware.gms.GroupHandler;
import sensei.middleware.gms.Message;
import sensei.middleware.domains.DomainMessage;
import sensei.middleware.domains.DomainsNarrower;
import sensei.middleware.domains.EndOfTransactionMessage;
import sensei.middleware.domains.MemberStateException;
import sensei.middleware.domains.Monitor;
import sensei.middleware.domains.MonitorException;
import sensei.middleware.domains.TransactionException;

import sensei.util.SenseiError;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * Class supporting the use of transactions
 */
public class TransactionsManager {

	//*************************************************************************************
	//**************************** CONSTRUCTOR ********************************************
	//*************************************************************************************

	/**
	 * Constructor requires the EventsQueue
	 */
	public TransactionsManager(int memberId, GroupHandler groupHandler, DomainMessagesHandler domainMessagesHandler, DomainsFactory messenger, EventsQueue eventsQueue, EventsQueueUser messagesConsumer) {
		Trace.code("TransactionsManager.java -> TransactionsManager ( int memberId , GroupHandler groupHandler , DomainMessagesHandler domainMessagesHandler , DomainsFactory messenger , EventsQueue eventsQueue , EventsQueueUser messagesConsumer )");
		this.memberId = memberId;
		this.groupHandler = groupHandler;
		this.domainMessagesHandler = domainMessagesHandler;
		this.eventsQueue = eventsQueue;
		this.messenger = messenger;
		this.messagesConsumer = messagesConsumer;
		transactionMonitors = new ArrayList();
		queuedMessages = new ArrayList();
		state = false;
	}

	//*************************************************************************************
	//**************************** FREE RESOURCES *****************************************
	//*************************************************************************************
	/**
	 * Frees any resources, the object can not be used anymore
	 */
	public void freeResources() {
		Trace.code("TransactionsManager.java -> void freeResources ( )");
		if (queuedMessages != null) {
			queuedMessages.clear();
			transactionMonitors.clear();
			transactionMonitors = null;
			groupHandler = null;
			domainMessagesHandler = null;
			eventsQueue = null;
			messagesConsumer = null;
			messenger = null;
			queuedMessages = null;
			state = false;
		}
	}

	public void becameStatefull() {
		Trace.code("TransactionsManager.java -> void becameStatefull ( )");
		state = true;
	}

	public boolean isOnTransaction() {
		Trace.code("TransactionsManager.java -> boolean isOnTransaction ( )");
		return !transactionMonitors.isEmpty();
	}

	public void startTransaction(Monitor monitor) throws MemberStateException, MonitorException {
		Trace.code("TransactionsManager.java -> void startTransaction ( Monitor monitor )");
		try {
			monitor.lock();
			synchronized (this) {
				if (transactionMonitors.isEmpty())
					domainMessagesHandler.startTransaction(this);
				transactionMonitors.add(monitor);
			}
		}
		catch (MemberStateException ex) {
			throw ex;
		}
		catch (Exception ex) { //RemoteException or CorbaException
			if (DomainsNarrower.isObjectNotExistException(ex))
				throw new MonitorException(ex.toString());
			else
				SenseiError.unhandledException(Consts.AREA, ex);
		}
	}

	public synchronized void endTransaction() throws MemberStateException, MonitorException, TransactionException {
		Trace.code("TransactionsManager.java -> synchronized void endTransaction ( )");
		int size = transactionMonitors.size();
		if (size == 0)
			throw new TransactionException();
		else {
			try {
				Monitor monitor = (Monitor) transactionMonitors.remove(--size);
				monitor.unlock();
				if (size == 0)
					transactionCompleted();
			}
			catch (MemberStateException ex) {
				throw ex;
			}
			catch (MonitorException ex) {
				throw ex;
			}
			catch (IOException ex) {
				if (DomainsNarrower.isObjectNotExistException(ex))
					throw new MonitorException(ex.toString());
				else
					SenseiError.unhandledException(Consts.AREA, ex);
			}
		}
	}

	void transactionCompleted() throws IOException {
		Trace.code("TransactionsManager.java -> void transactionCompleted ( )");
		domainMessagesHandler.endTransaction();
		//send a message with the queued messages

		int size = queuedMessages.size();
		if (size > 0) {
			int i = 0;
			Message[] messages = new Message[size];
			Iterator it = queuedMessages.iterator();
			while (it.hasNext()) {
				assert i < size;
				messages[i++] = (Message) it.next();
			}
			queuedMessages.clear();
			messenger.castEndOfTransaction(messages);
		}
	}

	public synchronized boolean castMessage(Message message) throws IOException {
		Trace.code("TransactionsManager.java -> synchronized boolean castMessage ( Message message )");
		//messages coming here are all transactionable
		if (isOnTransaction()) {
			//simulate that is sent, and store it on a queue to be sent later
			eventsQueue.newTransactionMessage(memberId, message);
			queuedMessages.add(message);
		}
		else
			groupHandler.castMessage(message);
		return true;
	}

	public void processMessageEndOfTransaction(int sender, EndOfTransactionMessage message) {
		Trace.code("TransactionsManager.java -> void processMessageEndOfTransaction ( int sender , EndOfTransactionMessage message )");
		//messages sent by this member are not processed
		if (sender != memberId) {
			//the rest of messages are processed sequentially.
			//if this member has no state, only the unqueuedOnST messages must be processed
			Message messages[] = message.actions;
			int size = messages.length;
			for (int i = 0; i < size; i++)
				if (processMessageOnEndOfTransaction(messages[i]))
					messagesConsumer.processCastMessageEvent(sender, messages[i]);
		}
	}

	boolean processMessageOnEndOfTransaction(Message message) {
		Trace.code("TransactionsManager.java -> boolean processMessageOnEndOfTransaction ( Message message )");
		return state || ((message instanceof DomainMessage) && (((DomainMessage) message).unqueuedOnST));
	}

	//*************************************************************************************
	//**************************** DATA MEMBERS *******************************************
	//*************************************************************************************

	List transactionMonitors; //monitor on each transaction
	int memberId;
	boolean state;
	List queuedMessages; //[Message]
	DomainMessagesHandler domainMessagesHandler;
	DomainsFactory messenger;
	GroupHandler groupHandler;
	EventsQueue eventsQueue;
	EventsQueueUser messagesConsumer;
}
