/**
  * File: ExitHandler.java
  * Content: Handles the correct exit from the group
  * Author: LuisM Pena
  * Date: 15th December 2001
  * Version: 0.51.00
  * Last change:
  *
  **/

package sensei.domains.utils;

import sensei.middleware.gms.GroupHandler;
import sensei.middleware.domains.DomainExpulsionReason;

import java.io.IOException;
import java.util.Timer;
import java.util.TimerTask;

/**
 * Class handling the correct exit from the group.
 * It covers several scenarios:
 * 1- When a member is already belonging to the group, and it is requested to leave it. In this case,
 *    when the exclusion event is received, a real exit is performed. It is possible to use a Timer for this
 *    scenario, to specify the maximum time to wait for the exclusion
 * 2- If the member is not yet belonging to the group, a real exit is directly performed.
 * 3- If the member is expulsed, leave also perform a real exit directly
 *
 * This class must received the exclusion events!
 */
public class ExitHandler extends TimerTask {

	public ExitHandler(ExitHandlerUser user) {
		this.user = user;
	}

	/**
	 * initExitProcedure is the initial point to perform a correct exit for group members. It is
	 * translated into a leave request, if the member is already belonging to a group.
	 * It is possible to specify the maximum time to wait for the exclusion; if the period lasts
	 * without having received an exclusion event, the user still receives a performExit call,
	 * but with the associated parameter set to false
	 */
	public void initExitProcedure(int maxMilliseconds) throws IOException {
		if (handler == null) {
			user.performExit(true);
		}
		else {
			leaving = true;
			left = false;
			timer = new Timer(false);
			timer.schedule(this, maxMilliseconds);
			handler.leaveGroup();
		}
	}

	public void run() {
		boolean goon;
		synchronized (this) {
			goon = leaving && !left;
			leaving = false;
			left = true;
		}
		if (goon) {
			user.performExit(false);
		}
	}

	/**
	 * When accepted in group, this method must be invoked with the correct group handler
	 */
	public void acceptedInGroup(GroupHandler handler) {
		this.handler = handler;
	}

	/**
	 * When excluded from the group, this method must be invoked with the given reason
	 */
	public void excludedFromGroup(DomainExpulsionReason reason) {
		boolean goon, alreadyLeft;
		synchronized (this) {
			alreadyLeft = left;
			goon = leaving;
			leaving = false;
			left = true;
		}
		if (!alreadyLeft) {
			if (goon && (handler != null)) {
				timer.cancel();
				user.performExit(true);
			}
			else {
				user.unsolicitedExclusion(reason);
			}
		}
	}

	ExitHandlerUser user;
	GroupHandler handler;
	Timer timer;
	boolean leaving, left;
}
