/**
  * File: Timer.java
  * Content: definition of a timer (not an accurate one)
  * Author: LuisM Pena
  * Date: 19th July 2000
  * Last change:
  * Version: 0.97.00
  *
  * Algorithm comentaries: The timer doesn't have to be accurate, and it just
  *		divides the timing in several periods (by default in 10), going into a
  *		loop where it sleeps n times. If the timer is restarted, the loop ends immediately
  *		without notifying the listener (to simplify, it accepts just one listener)
  *		and the loop starts again.
  *		This means that, if the previous 'sleep' had just started, the next timing
  *		will late for one period more (this is the accuracy factor).
  *		It's also important not to block any call. For example, if stop is called,
  *		a possibility would be to block it until the current sleep awakes and then
  *		unblock it. In this case, the next start() would take exactly the time requested.
  *		Because it's not blocked, the algorithm must support the case of, let's say,
  *		a stop/start or even a stop/start/stop.
  *		Independently of the accuracy taken by the algorithm, the sleep method is
  * 	neither accurate by himself, and the loops takes some (probably not needed
  *		to be considered) time, therefore don't expect good timings in the order
  *		of the millisecond.
  *
  **/

package sensei.util;

/**
  * Unaccurate timer based on the sleep method.
  * It allows the basic operations start() and stop() and also restart()
  **/
public class Timer implements Runnable {

	/**
	  * The timer needs to know the time to measure (in milliseconds)
	  * To release the timer, method release must be called. The timer doesn't start
	  * automatically, start() method must be called
	  * @param time the time to measure
	  * @param accuracy The accuracy needed, if it's zero or negative,
		  *      	the timer automatically calculates one
	  **/
	public Timer(TimerListener listener, long time, long accuracy) {
		this.listener = listener;
		setValues(time, accuracy);

		timerRunning = true;
		timerStopped = timerStarted = timerRestarted = false;

		associatedThread = new Thread(this);
		associatedThread.start();
	}

	/**
	  * Starts the timer, if it has been not stopped and can be started
	  * (it's not yet started).
	  **/
	public synchronized void start() {
		boolean notify = false;

		assert timerRunning;

		if (!timerStarted)
			//if not yet started, start it!
			timerStarted = notify = true;
		else if (timerStopped)
			//if it's started, and there is not any pending stop, it can't be started again!
			//otherwise, the current timer must finish, and the timer restarted.
			timerRestarted = notify = true;

		if (notify)
			notifyAll();
	}

	/**
	  * Stops the timer (doesn't free it)
	  * It can't be stopped if there have been problems with the timer or it's
	  * not started
	  **/
	public synchronized void stop() {
		assert timerRunning;

		//timer can only be stopped if it's already started
		if (timerStarted) {
			//if timer is already intended to be stopped, it could be still in 'restart'
			//mode, in that case, just abort the restart. If it was not planned to be
			//stopped, just stop it!
			timerStopped = true;
			timerRestarted = false;
		}
	}

	/**
	  * Restarts the timer, is just equivalent to do a stop() followed by a start()
	  **/
	public synchronized void restart() {
		stop();
		start();
	}

	/**
	  * Releases the timer, that can't be used anymore. This call blocks until the timer
	  * is effectively released
	  **/
	public void release() {
		Thread toRelease = associatedThread;
		if (toRelease != null) {
			synchronized (this) {
				timerStopped = true;
				timerRunning = false;
				notifyAll();
			}
			try {
				toRelease.join();
			}
			catch (Exception ex) {
				SenseiError.unhandledException(Consts.AREA_NAME, ex);
			}
		}
	}

	/**
	  * Runnable method, not to be called directly
	  **/
	public void run() {
		try {
			while (timerRunning) {
				//not start until timerStarted is set or timerRunning is unset
				synchronized (this) {
					while (!timerStarted && timerRunning)
						wait();
				}

				if (timerRunning) {
					//Sleep loops times for loopSleep time, breaking if timerStopped is set to true
					//Note that ends also sets timerStopped to true
					for (long i = 0; !timerStopped && (i < loops); i++)
						Thread.sleep(loopSleep);

					//And sleep a last time for the remaining time
					if (!timerStopped && finalSleep > 0)
						Thread.sleep(finalSleep);

					//if  timer was not stopped, notify to the listener
					if (!timerStopped)
						timerRunning = listener.timerExpired();

				}

				synchronized (this) {
					//Before starting a new loop, set the variables to the right values
					//If the timer is to be restarted, just make timeStarted equal to true,
					//otherwise just false, like the another variables
					timerStarted = timerRestarted;
					timerStopped = timerRestarted = false;
				}
			}
		}
		catch (Exception ex) {
			timerRunning = false;
			SenseiError.unhandledException(Consts.AREA_NAME, ex);
		}

		//free resources
		listener = null;
		associatedThread = null;
	}

	void setValues(long time, long accuracy) {
		if (time <= 0) {
			loops = 0;
			loopSleep = finalSleep = 0;
		}
		else {
			if (accuracy <= 0)
				accuracy = 1 + time / accuracyPeriods;
			loops = time / accuracy;
			finalSleep = time % accuracy;
			loopSleep = accuracy;
		}
	}

	long loops;
	long loopSleep;
	long finalSleep;

	boolean timerRunning; //timer is alive
	boolean timerRestarted; //start called with a stop pending
	boolean timerStarted; //timer must start (or it's started)
	boolean timerStopped; //timer must stop (or it's being stopped)

	TimerListener listener;
	Thread associatedThread;

	static final long accuracyPeriods = 10;
}
