/**
  * File: Logger.java
  * Content: Basic logging facilities
  * Author: LuisM Pena
  * Date: 1st October 2000
  * Version: 0.97.00
  * Last change:
  *
  *	Comments:
  *		Logger includes the functionality to log several parts of the system.
  *		The current implementation dumps all the logging to the same file
  *
  *   Functionality is added to dump the information to a socket. In that
  *   case, the LoggerServer receives the information, dumping it any way
  *   to the file using the same classes.
  *
  *   The main problem on the Logging is the configuration. Which socket to
  *   use? Or which name must be used for the logging file? It cannot use the
  *   Configuration class, as the Configuration logs itself all the information.
  *
  *   The answer is to let the system be lazy configured. The logging is initially
  *   stored in memory and when the socket is choosen the information is dumped
  *   there. If local logging is choosen, a file stores instead the logging.
  *
  *   The problem with this approach is to be too lazy, as the memory to allocate
  *   can be quite big. Nevertheless, the Configuration class and some other classes
  *   make quick use of the Logging and therefore, at least when used under sensei,
  *   this problems should not bother much.
  *
  *   If the system shutdowns before selecting a file or socket, the information is
  *   directly dumped into the output stream from the MemoryLogStream
  **/

package sensei.util.logging;

import sensei.util.SenseiError;

import java.util.HashMap;

public class Logger
{
  //*************************************************************************************//
  //**************************** GET LOGGER *********************************************//
  //*************************************************************************************//

  /**
    * Returns a Logger instance to be used for the specified part. Returns null if confirm
    * is not true
    * @param confirm must be set to true to obtain a Logger
    **/
  static public Logger getLogger(String loggingPart, boolean confirm)
  {
    Logger ret=null;
    if (confirm)
      synchronized (loggers)
      {
        //it's first tried directly if it's already created such a Logger
        ret = (Logger)(loggers.get(loggingPart));

        if (ret==null)
        {
          if (commonWriter==null)
          {
            commonWriter = initialWriter = new MemoryLogStream();
          }

          ret = new Logger(loggingPart);
          loggers.put(loggingPart, ret);
        }
      }
    return ret;
  }

  //*************************************************************************************//
  //**************************** SET LOG NAME *******************************************//
  //*************************************************************************************//

  /**
    * Loggers are created with temporal names. This method defines the final name for
    * the files. It is not verified if the method is called more than once
    **/
  static public void setLogName(String name)
  {
    synchronized(loggers)
    {
      if (initialWriter==null)
        commonWriter.setLogName(name);
      else
        initialWriter.setLogName(name);
    }
  }

  //*************************************************************************************//
  //**************************** USE REMOTE LOGGING *************************************//
  //*************************************************************************************//

  /**
    * Informs the Logger to use remote logging, specifying the port to use.
    * This port can only be set once, and only if useLocalLogging has not been
    * called. Otherwise, the call is just discarded
    **/
  static public void useRemoteLogging(int port)
  {
    synchronized(loggers)
    {
      if (initialWriter==null)
        SenseiError.error(LOGGING, "Calling Logger::useRemoteLogging when the mode is already set");
      else
      {
        commonWriter = new SocketLogStream(port);
        initialWriter.dumpToLogStream(commonWriter);
        initialWriter=null;
        includeTime=false; //done on the remote logging
      }
    }
  }

  //*************************************************************************************//
  //**************************** USE LOCAL LOGGING **************************************//
  //*************************************************************************************//

  /**
    * Informs the Logger to use local logging. This call is discarded if UseRemoteLogging
    * has been alreadz called
    **/
  static public void useLocalLogging()
  {
    synchronized(loggers)
    {
      if (initialWriter==null)
        SenseiError.error(LOGGING, "Calling Logger::useLocalLogging when the mode is already set");
      else
      {
        commonWriter = new FileLogStream(true);
        initialWriter.dumpToLogStream(commonWriter);
        initialWriter=null;
      }
    }
  }

  //*************************************************************************************//
  //**************************** DO NOT USE LOGGING *************************************//
  //*************************************************************************************//

  /**
    * Informs the Logger to use no logging. This call is discarded if UseRemoteLogging or
    * UseLocalLogging has been already called
    **/
  static public void doNotUseLogging()
  {
    synchronized(loggers)
    {
      if (initialWriter==null)
        SenseiError.error(LOGGING, "Calling Logger::doNotUseLogging when the mode is already set");
      else
      {
        initialWriter.close();
        initialWriter=null;
      }
    }
  }

  //*************************************************************************************//
  //**************************** CONSTRUCTOR ********************************************//
  //*************************************************************************************//

  /**
    * It is not possible to create Logger classes outside the singleton factory
    * @param loggingPart the area to log
    **/
  private Logger(String loggingPart)
  {
    this.loggingPart = loggingPart;
    noAutoLogging = loggingPart != LOGGING;
    string = new StringBuffer();
  }

  //*************************************************************************************//
  //**************************** LOG ****************************************************//
  //*************************************************************************************//

  /**
    * Logs the given message
    **/
  public void log(String message)
  {
    if (doLogging)
    {
      synchronized (loggers)
      {
        if (doLogging)
        {
          try
          {
            if (noAutoLogging && commonWriter != null)
            {
              string.delete(0,Integer.MAX_VALUE);
              if (includeTime)
                string.append(String.valueOf(System.currentTimeMillis() - initialTime));
              string.append(" : ");
              string.append(loggingPart);
              string.append(" : ");
              string.append(message);
              commonWriter.println(string.toString());
            }
            else
              System.out.println(message);
          }
          catch(Exception ex)
          {
            SenseiError.unhandledException(LOGGING, ex);
          }
        }
      }
    }
  }

  //*************************************************************************************//
  //**************************** DATA MEMBERS *******************************************//
  //*************************************************************************************//

  //individual variables
  final String loggingPart;
  final boolean noAutoLogging;
  StringBuffer string;

  //*************************************************************************************//
  //**************************** DATA MEMBERS *******************************************//
  //*************************************************************************************//

  static HashMap loggers = new HashMap();		//existing loggers
  static LogStream commonWriter = null;			//needed if one file will include every logging
  static MemoryLogStream initialWriter=null;
  static boolean includeTime=true, doLogging=true;

  static final String LOGGING        = "sensei.util.Logging";
  static final long initialTime      = System.currentTimeMillis();
}
